import sys
import logging
import transaction
from plone import api

from zope.component.hooks import getSite
from zope.lifecycleevent import modified
from smtplib import SMTPException

from Products.CMFCore.utils import getToolByName
from plone.app.contentlisting.interfaces import IContentListing
from Products.CMFCore.interfaces import IFolderish
from collective.easyform.interfaces import ISaveData
from collective.easyform.api import get_schema
from collective.easyform.api import getFieldsInOrder

from sandia.conferencepolicy.content.utillib import get_attendees_for_conference, registration_data_for_attendee
from datetime import datetime

# LOGGER = logging.getLogger(__name__)
LOGGER = logging.getLogger("Plone")
ATTENDEE_SAVE_DATA_ADAPTER_TITLE = 'Attendee Emails – DO NOT REMOVE'
ALTERNATE_ATTENDEE_SAVE_DATA_ADAPTER_TITLE = 'Participants'

ATTENDEE_SAVE_DATA_ADAPTERS = [
    ATTENDEE_SAVE_DATA_ADAPTER_TITLE,
    ALTERNATE_ATTENDEE_SAVE_DATA_ADAPTER_TITLE
]

def get_all_sites(app_object):
    return [app_object[next_object['id']] for next_object in app_object.__dict__["_objects"] if "meta_type" in next_object and next_object["meta_type"] == "Plone Site"]

def hasWorkflow(obj):
    """
    Return boolean, indicating whether obj has a workflow assigned, or not.
    """
    return len(obj.portal_workflow.getWorkflowsFor(obj)) > 0

def getState(obj):
    """
    Return workflow-state-id or None, if no workflow is assigned.
    Show possible error on the console and log it.
    """
    if hasWorkflow(obj):
        try: return obj.portal_workflow.getInfoFor(obj, 'review_state')
        except ExceptionError as err: obj.plone_log(err)
    else: return None

def display_all_content(root_object, prefix=""):

    def recurse(context, prefix=""):
        """ Recurse through all content within passed context """

        context_state = getState(context)
        print(f"{prefix} {context.Title()} - {context.Type()} Workflow State:{getState(context)}")

        # Make sure that we recurse to real folders only,
        # otherwise contentItems() might be acquired from higher level
        if IFolderish.providedBy(context):
            for id, item in context.contentItems():
                recurse(item, prefix=f"\t{prefix}",)

    recurse(root_object, prefix=prefix)

def get_all_items_of_type(root_object, type_string):

    def recurse_all_items_of_type(context, type_string):
        """ Recurse through all content within passed context """
        items_list = []
        if context.Type() == type_string:
            items_list.append(context)

        # Make sure that we recurse to real folders only,
        # otherwise contentItems() might be acquired from higher level
        if IFolderish.providedBy(context):
            try:
                for id, item in context.contentItems():
                    items_list.extend(recurse_all_items_of_type(item, type_string))
            except Exception as ignoreException:
                pass
        return items_list

    return recurse_all_items_of_type(root_object, type_string)

def get_all_groups_for_site(site_object):
    return get_all_items_of_type(site_object, "Group")

def get_all_conferences_for_site(site_object):
    return get_all_items_of_type(site_object, "Conference")

def get_notifications_folder_list_for_conference(conference_object):
    # display_all_content(conference_object)
    notifications_folder_list = [folder for folder in get_all_items_of_type(conference_object, "Folder") if folder.Title() == 'Notifications']
    if len(notifications_folder_list) == 0:
        LOGGER.warning(f"Conference: {conference_object.Title()} has no Notifications Folder!")
    elif (len(notifications_folder_list) > 1):
        LOGGER.warning(f"Conference: {conference_object.Title()} has {len(notifications_folder_list)} Notifications Folders - something is not right here!")
    
    return notifications_folder_list

def get_notifications_from_folder(notifications_folder):
    # display_all_content(notifications_folder, prefix="\t\t\t")
    return get_all_items_of_type(notifications_folder, "Notification")

def get_notifications_for_conference(conference_object):
    conference_notifications_folder_list = get_notifications_folder_list_for_conference(conference_object)
    conference_notifications = []
    if len(conference_notifications_folder_list) > 0:
        for next_notifications_folder in conference_notifications_folder_list:
            notifications_from_folder = get_notifications_from_folder(next_notifications_folder)
            conference_notifications.extend(notifications_from_folder)
    return conference_notifications

def orig_get_attendees_for_conference(conference_object):
    """return a list of the attendees"""
    all_email_addresses = []
    all_attendees = []
    # read all registration forms in this conference
    brains = api.content.find(context=conference_object, portal_type='EasyForm')
    attendee_keys = None
    for brain in brains:
        form = brain.getObject()
        # get all the form's save data adapters
        all_save_data_adapters = form.restrictedTraverse("@@get_save_data_adapters")
        results = all_save_data_adapters()
        for saver in results:
            # must have the ID 'attendee_emails___do_not_remove'
            if saver.title in ATTENDEE_SAVE_DATA_ADAPTERS:
                save_data_adapter = ISaveData(saver)
                data = save_data_adapter.getSavedFormInput()
                # relying on the email address field to have the ID 'replyto'
                if len(data) > 0:
                    for item in data:
                        if not attendee_keys or len(item.keys()) > len(attendee_keys):
                            attendee_keys = item.keys()

                        approved_value = 'No'
                        if 'approved' in item:
                            approved_value = item['approved']
                        checked_in_value = 'No'
                        if 'checked_in' in item:
                            checked_in_value = item['checked_in']
                        replacing_attendee_value = ''
                        if 'replacing_attendee' in item:
                            replacing_attendee_value = item['replacing_attendee']
                        
                        item_as_tuple = map(lambda x: frozenset(x) if isinstance(x, set) else x, tuple([item[next_key] if next_key in item else '' for next_key in attendee_keys]))
                        all_attendees.append( item_as_tuple )

    unique_attendees_set = set(all_attendees)
    unique_attendees_list = [dict(zip(attendee_keys, x)) for x in unique_attendees_set]
    return unique_attendees_list 

def text_to_html(a_text):
    return "<br/>".join(a_text.split("\n"))

def send_formatted_notification(mailhost, mail_sent_from, notification, attendees_dict):
    for next_attendee in attendees_dict:
        attendee_registration_details = registration_data_for_attendee(attendees_dict[next_attendee])
        try:
            mailhost.send(
                notification.body.raw.replace('$ATTENDEE_SUBMISSION_INFO', text_to_html(attendee_registration_details)), 
                next_attendee, 
                mail_sent_from, 
                subject=notification.email_subject, 
                charset="utf-8", 
                msg_type='text/html',
            )
        except SMTPException:
            LOGGER.error('SMTP exception while trying to send an ' +
                         'email from %s to %s',
                         mail_sent_from,
                         next_attendee)
        # message_display = "\n"
        # message_display = message_display + "\n\tSend Email:"
        # message_display = message_display + f"\n\t\tSubject: {notification.email_subject}"
        # message_display = message_display + f"\n\t\tBody: {notification.body.raw.replace('$ATTENDEE_SUBMISSION_INFO', attendee_registration_details)}"
        # message_display = message_display + "\n"
        # message_display = message_display + f"\n\t\tDate to Send: {notification.send_date}"
        # message_display = message_display + f"\n\t\tNotification has been sent?: {notification.sent_flag}"
        # message_display = message_display + "\n" + ("*" * 100)
        # LOGGER.info(message_display)

def process_notifications():
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)

    LOGGER.info("Notifiations Processor starting up . . .")
    portal = api.portal.get()
    app = portal.restrictedTraverse('/')
    my_all_sites = get_all_sites(app)
    for next_site in my_all_sites:
        mailhost = getToolByName(next_site, 'MailHost')
        LOGGER.info(f"Mail host: {mailhost}")

        LOGGER.info(f"Processing Site: '{next_site.Title()}'")
        for next_conference in get_all_conferences_for_site(next_site):
            conference_display = "\n"
            conference_display = conference_display + f"\tConference: {next_conference.Title()}"
            mail_sent_from = next_conference.contact_email or 'noreply@sandia.gov'
            LOGGER.info(f"Send mail from: {mail_sent_from}")
            attendees = get_attendees_for_conference(next_conference)
            conference_display = conference_display + "\n\tAttendees:"
            for next_attendee in attendees:
                conference_display = conference_display + f"\n\t\t{next_attendee}: {len(attendees[next_attendee])}"

            next_conference_notifications_list = get_notifications_for_conference(next_conference)
            LOGGER.info(conference_display)
            if len(next_conference_notifications_list) > 0:
                for next_notification in next_conference_notifications_list:
                    existing_value = getattr(next_notification, "sent_flag")
                    if not next_notification.sent_flag and datetime.now().date() > next_notification.send_date: 
                        send_formatted_notification(mailhost, mail_sent_from, next_notification, attendees)
                        next_notification.sent_flag = True
                        transaction.commit()
                        existing_value = getattr(next_notification, "sent_flag")
                        modified(next_notification)                        

                    else:
                        if next_notification.sent_flag:
                            LOGGER.info(f"Notification {next_notification.Title()} for Conference {next_conference.Title()} has already been sent . . .")
                        else:
                            LOGGER.info(f"Notification {next_notification.Title()} for Conference {next_conference.Title()} is not ready to be sent yet . . .")
            else:
                LOGGER.info(f"Conference {next_conference.Title()} has no notifications configured . . .")
        LOGGER.info(f"Processing of Site: '{next_site.Title()}' complete . . .")
    LOGGER.info("Notifications Processor complete.")
