from datetime import datetime
import logging
from plone import api
from dataclasses import dataclass

from zope.interface import provider
from plone.supermodel.interfaces import IDefaultFactory
from zope.schema.interfaces import IContextAwareDefaultFactory
from Products.CMFCore.interfaces import ISiteRoot
from collective.easyform.interfaces import ISaveData
from Products.CMFCore.utils import getToolByName
from plone.app.contentlisting.interfaces import IContentListing
from Products.CMFCore.interfaces import IFolderish

LOGGER = logging.getLogger(__name__)
NOTIFICATION_EVENT_DATE = "DATE_HERE"
NOTIFICATION_EVENT_TIME = "TIME_HERE"
NOTIFICATION_EVENT_NAME = "TTX"
NOTIFICATION_EVENT_ORGANIZER = "ORGANIZER_HERE"
NOTIFICATION_EVENT_ORGANIZER_EMAIL = "ORGANIZER_EMAIL_HERE"
NOTIFICATION_EMAIL_BODY_TEMPLATE = """
<p>This is a reminder that you are registered to participate in TTX Tabletop<br />
Exercise on <strong>DATE_HERE</strong> from <strong>TIME_HERE</strong>. If you have any questions or concerns,<br />
please contact <strong>ORGANIZER_HERE</strong> via email at <strong>ORGANIZER_EMAIL_HERE</strong>.  The<br />information you submitted is below for your records.</p>
<p></p>
<p>$ATTENDEE_SUBMISSION_INFO</p>
<p></p>
"""

ATTENDEE_SAVE_DATA_ADAPTER_TITLE = 'Attendee Emails – DO NOT REMOVE'
ALTERNATE_ATTENDEE_SAVE_DATA_ADAPTER_TITLE = 'Participants'

ATTENDEE_SAVE_DATA_ADAPTERS = [
    ATTENDEE_SAVE_DATA_ADAPTER_TITLE,
    ALTERNATE_ATTENDEE_SAVE_DATA_ADAPTER_TITLE
]

CONFERENCE_CERTIFICATE_IMAGE_DOCUMENT_NAME = "Conference Certificate Image"

@dataclass
class InvalidConferenceItem:
    title: str
    start: datetime.date
    contact_name: str
    contact_email: str

def get_all_items_of_type(root_object, type_string):

    def recurse_all_items_of_type(context, type_string):
        """ Recurse through all content within passed context """
        items_list = []
        if context.Type() == type_string:
            items_list.append(context)

        # Make sure that we recurse to real folders only,
        # otherwise contentItems() might be acquired from higher level
        if IFolderish.providedBy(context):
            try:
                for id, item in context.contentItems():
                    items_list.extend(recurse_all_items_of_type(item, type_string))
            except Exception as ignoreException:
                pass
        return items_list

    return recurse_all_items_of_type(root_object, type_string)

def get_documents_from_folder(documents_folder):
    return get_all_items_of_type(documents_folder, "File")

def get_folder_list_for_conference(conference_object, folder_title):
    # display_all_content(conference_object)
    folder_list = [folder for folder in get_all_items_of_type(conference_object, "Folder") if folder.Title() == folder_title]
    if len(folder_list) == 0:
        LOGGER.warning(f"Conference: {conference_object.Title()} has no {folder_title} Folder!")
    elif (len(folder_list) > 1):
        LOGGER.warning(f"Conference: {conference_object.Title()} has {len(folder_list)} {folder_title} Folders - something is not right here!")
    
    return folder_list

def get_documents_for_conference(conference_object):
    conference_documents_folder_list = get_folder_list_for_conference(conference_object, "Documents")
    conference_documents = []
    if len(conference_documents_folder_list) > 0:
        for next_documents_folder in conference_documents_folder_list:
            documents_from_folder = get_documents_from_folder(next_documents_folder)
            conference_documents.extend(documents_from_folder)
    return conference_documents

def get_certificate_image_for_conference(conference_object):
    conference_documents = get_documents_for_conference(conference_object)
    result_document = None
    for next_document in conference_documents:
        if next_document.Title().lower() == CONFERENCE_CERTIFICATE_IMAGE_DOCUMENT_NAME.lower():
            result_document = next_document
    return result_document

def getConferenceFromContext(object):
    my_chain_to_root = list(getAcquisitionChain(object))
    # chain[0] = the CertificateMailer
    # chain[1] = the Conference
    # chain[2] = the Site Root
    if len(my_chain_to_root) < 3 or my_chain_to_root[1].Type() != "Conference":
        return None
    
    my_conference_object = my_chain_to_root[1]
    return my_conference_object

def getAcquisitionChain(object):
    """
    @return: List of objects from context, its parents to the portal root

    Example::

        chain = getAcquisitionChain(self.context)
        print "I will look up objects:" + str(list(chain))

    @param object: Any content object
    @return: Iterable of all parents from the direct parent to the site root
    """

    # It is important to use inner to bootstrap the traverse,
    # or otherwise we might get surprising parents
    # E.g. the context of the view has the view as the parent
    # unless inner is used
    inner = object.aq_inner

    iter = inner

    while iter is not None:
        yield iter

        if ISiteRoot.providedBy(iter):
           break

        if not hasattr(iter, "aq_parent"):
            raise RuntimeError("Parent traversing interrupted by object: " + str(parent))

        iter = iter.aq_parent

def formattedDate(a_date):
    return a_date.strftime('%B %d, %Y')

def formattedTime(a_date):
    return a_date.strftime('%H:%M')

def getConferenceFromNotificationContext(object):
    my_chain_to_root = list(getAcquisitionChain(object))
    # chain[0] = the Notifications Folder
    # chain[1] = the Conference
    # chain[2] = the Site Root
    if len(my_chain_to_root) < 3 or my_chain_to_root[1].Type() != "Conference":
        return InvalidConferenceItem(title="Warning: It looks like you are attempting to add a notification outside of a Notifications Folder . . .",
        start=datetime.now(),
        contact_name="",
        contact_email="",)
    
    my_conference_object = my_chain_to_root[1]
    return my_conference_object

@provider(IDefaultFactory)
def notificationEmailSubject1():
    return "TTX Tabletop Exercise"

@provider(IContextAwareDefaultFactory)
def notificationEmailSubject(context):
    my_conference_object = getConferenceFromNotificationContext(context)
    if not isinstance(my_conference_object, InvalidConferenceItem):    
        return f" {my_conference_object.title} Tabletop Exercise on {formattedDate(my_conference_object.start)}"
    
    return ''

@provider(IContextAwareDefaultFactory)
def notificationEmailBody(context):
    my_conference_object = getConferenceFromNotificationContext(context)
    my_conference_organizer = my_conference_object.contact_name or ''
    my_conference_organizer_email = my_conference_object.contact_email or ''

    return NOTIFICATION_EMAIL_BODY_TEMPLATE.replace(
            NOTIFICATION_EVENT_DATE, formattedDate(my_conference_object.start)
        ).replace(
            NOTIFICATION_EVENT_TIME, formattedTime(my_conference_object.start)
        ).replace(
            NOTIFICATION_EVENT_NAME, my_conference_object.title
        ).replace(
            NOTIFICATION_EVENT_ORGANIZER, my_conference_object.contact_name
        ).replace(
            NOTIFICATION_EVENT_ORGANIZER_EMAIL, my_conference_object.contact_email
        )

def get_attendees_for_conference(conference_object):
    """return a dictionary of the attendees - keyed by email"""
    attendees_dict = dict()
    # read all registration forms in this conference
    brains = api.content.find(context=conference_object, depth=6, portal_type='EasyForm')
    attendee_keys = None
    for brain in brains:
        form = brain.getObject()
        # get all the form's save data adapters
        all_save_data_adapters = form.restrictedTraverse("@@get_save_data_adapters")
        for saver in all_save_data_adapters():
            # must have the ID 'attendee_emails___do_not_remove'
            if saver.title in ATTENDEE_SAVE_DATA_ADAPTERS:
                save_data_adapter = ISaveData(saver)
                data = save_data_adapter.getSavedFormInput()
                # relying on the email address field to have the ID 'replyto'
                if len(data) > 0:
                    for item in data:
                        if not attendee_keys or len(item.keys()) > len(attendee_keys):
                            attendee_keys = item.keys()

                        approved_value = 'No'
                        if 'approved' not in item:
                            item['approved'] = 'No'
                        if 'checked_in' not in item:
                            item['checked_in'] = 'No'
                        if 'replacing_attendee' not in item:
                            item['replacing_attendee'] = ''
                        
                        # Add this attendee to the attendees dict
                        next_attendee_email = item['replyto']
                        if next_attendee_email not in attendees_dict:
                            attendees_dict[next_attendee_email] = dict()
                        attendees_dict[next_attendee_email][item['id']] = item
    return attendees_dict

def format_response(a_response):
    if isinstance(a_response, set):
        return ",".join([element for element in a_response])
    elif isinstance(a_response, bool):
        return "Yes" if a_response else "No"

    return a_response

HIDDEN_REGISTRATION_KEYS = ["recaptcha", "approved", "checked_in", "dt", "id", "ignore"]

def registration_data_for_attendee(attendee):
    latest_details_id = sorted(attendee.keys())[-1]
    attendee_details = attendee[latest_details_id]
    result = ""
    for key in attendee_details:
        if key == "replacing_attendee" and not attendee_details[key]:
            key = "ignore"

        if key == "replyto":
            display_key = "Email"
        else:
            display_key = key.replace('_', ' ').title()
        if key.lower() not in HIDDEN_REGISTRATION_KEYS:
            result += f"\n{display_key}: {format_response(attendee_details[key])}"
    
    return result